import base64
from PIL import Image
import io

import requests


def base64_to_image(base64_string, output_path):
    """
    This function converts a base64 string to an image.
    """
    image_data = base64.b64decode(base64_string)
    image = Image.open(io.BytesIO(image_data))
    image.save(output_path)


def image_to_base64(image_path):
    """
    This function converts an image to a base64 string.
    """
    with open(image_path, "rb") as image_file:
        encoded_string = base64.b64encode(image_file.read()).decode('utf-8')
    return encoded_string


def read_image(image_path):
    try:
        img = Image.open(image_path)
        return img
    except IOError:
        print("Unable to load image")
        return None


def load_image(url_or_path: str):
    name = url_or_path.removeprefix(
        "https://unsplash.com/photos/").split("/")[0]
    cache = read_image(f"{name}.jpg")

    if cache is not None:
      return cache

    if url_or_path.startswith("http://") or url_or_path.startswith("https://"):
        print(f"Downloading Images {name}")
        raw_image = requests.get(url_or_path, stream=True).raw
        image = Image.open(raw_image)
        image.save(f"{name}.jpg")
        return image
    else:
        return Image.open(url_or_path)



from typing import Dict
from pydantic import BaseModel


class FeaturesState(BaseModel):
    on_llm_start: bool = True
    on_llm_end: bool = True
    on_chat_start: bool = True
    on_chat_end: bool = True
    on_llm_response: bool = True
    on_llm_new_token: bool = True
    on_embeddings_start: bool = True
    on_embeddings_end: bool = True
    on_image_start: bool = True
    on_image_end: bool = True
    on_tts_start: bool = True
    on_tts_end: bool = True
    on_transcription_start: bool = True
    on_transcription_end: bool = True
    on_retry: bool = True
    
    @classmethod
    def only(cls, states: Dict[str, bool]):
        res = FeaturesState.none()
        for state in states.items():
            setattr(res, state[0], state[1])
        return res

    @classmethod
    def none(cls):
        res = FeaturesState(
            on_chat_end=False,
            on_chat_start=False,
            on_image_end=False,
            on_image_start=False,
            on_transcription_end=False,
            on_transcription_start=False,
            on_tts_end=False,
            on_tts_start=False,
            on_embeddings_end=False,
            on_embeddings_start=False,
            on_llm_end=False,
            on_llm_new_token=False,
            on_llm_response=False,
            on_llm_start=False,
            on_retry=False
        )
        return res




from .callbacks import CallbackHandler, CallbackManager, Callbacks
from .manipulations import ManipulationHandler, ManipulationManager, Manipulations

from .errors import (
    APIConnectionError,
    APIError,
    APIResponseValidationError,
    AuthenticationError,
    BadRequestError,
    BudgetExceededError,
    ContextWindowExceededError,
    OpenAIError
)

from .provider_base import (
    ProviderBase,
    completion_decorator,
    chat_decorator,
    embeddings_decorator,
    image_generation_decorator,
    tts_decorator,
    transcription_decorator,
)

from .utils import (
    create_base_retry_decorator,
)

from .provider_inputs import (
    BaseProviderInput,
    FireworksInput,
    ProviderType,
    GoogleGenAIInput,
    CohereInput,
    LlamacppInput,
    OpenAIInput,
    TogetherInput,
    CompleteInput,
    HuggingFaceInput,
    CTranslateInput,
    VoyageInput,
    GradientInput,
    NovitaInput,
    SegmindInput,
    TTSInput,
    FasterWhisperInput
)

from .response_types import (
    BaseMessage,
    Message,
    ChatMessages,
    HumanMessage,
    SystemMessage,
    AIMessage,
    UserMessage,
    CompletionData,
    ChatData,
    ChatCompletionRequest,
    ChatResponse,
    CompletionChoices,
    CompletionRequest,
    CompletionResponse,
    ChatContent,
    ChatChoices,
    Usage,
    FunctionCallContent,
    EmbeddingsData,
    ImageContent,
    ImageData,
    ImageRequest,
    ImageResponse,
    TTSData,
    TTSRequest,
    TTSResponse,
    TranscriptionData,
    TranscriptionRequest,
    TranscriptionResponse,
    VoiceSegments
)

from .base import (
    FeaturesState
)

__all__: list[str] = [
    'CallbackHandler',
    'CallbackManager',
    'Callbacks',
    'ManipulationHandler',
    'ManipulationManager',
    'Manipulations',
    'APIConnectionError',
    'APIError',
    'APIResponseValidationError',
    'AuthenticationError',
    'BadRequestError',
    'BudgetExceededError',
    'ContextWindowExceededError',
    'OpenAIError',
    'BaseMessage',
    'UserMessage',
    'HumanMessage',
    'SystemMessage',
    'AIMessage',
    'Message',
    'ChatMessages',
    'create_base_retry_decorator',
    'completion_decorator',
    'chat_decorator',
    'embeddings_decorator',
    'image_generation_decorator',
    'tts_decorator',
    'CompletionData',
    'ChatData',
    'ChatCompletionRequest',
    'CompletionChoices',
    'ChatContent',
    'ChatChoices',
    'ChatResponse',
    'CompletionRequest',
    'CompletionResponse',
    'FunctionCallContent',
    'Usage',
    'ProviderBase',
    'ProviderType',
    'BaseProviderInput',
    'FireworksInput',
    'GoogleGenAIInput',
    'CohereInput',
    'LlamacppInput',
    'OpenAIInput',
    'TogetherInput',
    'CompleteInput',
    'FeaturesState',
    'EmbeddingsData',
    'CTranslateInput',
    'HuggingFaceInput',
    'VoyageInput',
    'GradientInput',
    'NovitaInput',
    'ImageContent',
    'ImageData',
    'ImageRequest',
    'ImageResponse',
    'TTSData',
    'TTSRequest',
    'TTSResponse',
    'SegmindInput',
    'TTSInput',
    'TranscriptionData',
    'TranscriptionRequest',
    'TranscriptionResponse',
    'FasterWhisperInput',
    'transcription_decorator',
    'VoiceSegments',
]



import logging
import tiktoken
from typing import Any, Callable, List, Optional, Type
from tenacity import RetryCallState, before_sleep_log, retry, retry_if_exception_type, stop_after_attempt, wait_exponential

from .callbacks import CallbackManager
from ..utils import logger


def create_base_retry_decorator(
    error_types: List[Type[BaseException]],
    max_retries: int = 1,
    run_manager: Optional[CallbackManager] = None,
) -> Callable[[Any], Any]:
    """Create a retry decorator for a given LLM and provided list of error types."""

    _logging = before_sleep_log(logger, logging.WARNING)

    def _before_sleep(retry_state: RetryCallState) -> None:
        _logging(retry_state)
        if run_manager:
            run_manager.on_retry(retry_state, [])
        return None

    min_seconds = 4
    max_seconds = 10
    retry_instance = retry_if_exception_type(error_types[0]) if len(
        error_types) > 0 else retry_if_exception_type()
    for error in error_types[1:]:
        retry_instance = retry_instance | retry_if_exception_type(error)
    return retry(
        reraise=True,
        stop=stop_after_attempt(max_retries),
        wait=wait_exponential(multiplier=1, min=min_seconds, max=max_seconds),
        retry=retry_instance,
        before_sleep=_before_sleep,
    )


def num_tokens_from_string(string: str, encoding_name: str = "cl100k_base") -> int:
    """Returns the number of tokens in a text string."""
    encoding = tiktoken.get_encoding(encoding_name)
    num_tokens = len(encoding.encode(string))
    return num_tokens



from .integrations.callbacks.aim import AimCallbackHandler 
from .integrations.callbacks.confident import DeepEvalCallbackHandler
from .integrations.callbacks.context import ContextCallbackHandler
from .integrations.callbacks.launry import LunaryCallbackHandler
from .integrations.callbacks.prompt_layer import PromptLayerCallbackHandler
from .integrations.callbacks.streaming_stdout import StreamingStdOutCallbackHandler
from .integrations.manipulations.cache import CacheManipulationHandler
from .integrations.manipulations.prompt import PromptManipulationHandler
from .providers.openai import OpenAIProvider
from .providers.cohere import CohereProvider
from .providers.ctranslate import CTranslateProvider
from .providers.eleven_labs import ElevenLabsProvider
from .providers.fake import FakeProvider
from .providers.faster_whisper import FasterWhisperProvider
from .providers.fireworks import FireworksProvider
from .providers.genai import GoogleGenAIProvider
from .providers.gf4 import FreeProvider
from .providers.gradient import GradientProvider
from .providers.huggingface import HuggingFaceProvider
from .providers.llama_cpp import LlamacppProvider
from .providers.together import TogetherProvider
from .providers.novita import NovitaProvider
from .providers.segmind import SegmindProvider
from .providers.tts import TTSProvider
from .providers.voyage import VoyageProvider
from .types import *
from .utils import *


import re
import warnings
from typing import Any, Callable, Iterable, List, Optional

class TextSplitter():

    def __init__(
        self,
        chunk_size: int = 512,
        chunk_overlap: int = 100,
        length_function: Callable[[str], int] = len,
        keep_separator: bool = False,
        add_start_index: bool = False,
        strip_whitespace: bool = True,
    ) -> None:
        if chunk_overlap > chunk_size:
            raise ValueError(
                f"Got a larger chunk overlap ({chunk_overlap}) than chunk size "
                f"({chunk_size}), should be smaller."
            )
        self._chunk_size = chunk_size
        self._chunk_overlap = chunk_overlap
        self._length_function = length_function
        self._keep_separator = keep_separator
        self._add_start_index = add_start_index
        self._strip_whitespace = strip_whitespace

    def _join_docs(self, docs: List[str], separator: str) -> Optional[str]:
        text = separator.join(docs)
        if self._strip_whitespace:
            text = text.strip()
        if text == "":
            return None
        else:
            return text

    def _merge_splits(self, splits: Iterable[str], separator: str) -> List[str]:
        separator_len = self._length_function(separator)

        docs = []
        current_doc: List[str] = []
        total = 0
        for d in splits:
            _len = self._length_function(d)
            if (
                total + _len + (separator_len if len(current_doc) > 0 else 0)
                > self._chunk_size
            ):
                if total > self._chunk_size:
                    warnings.warn(
                        f"Created a chunk of size {total}, "
                        f"which is longer than the specified {self._chunk_size}"
                    )
                if len(current_doc) > 0:
                    doc = self._join_docs(current_doc, separator)
                    if doc is not None:
                        docs.append(doc)
                    # Keep on popping if:
                    # - we have a larger chunk than in the chunk overlap
                    # - or if we still have any chunks and the length is long
                    while total > self._chunk_overlap or (
                        total + _len +
                        (separator_len if len(current_doc) > 0 else 0)
                        > self._chunk_size
                        and total > 0
                    ):
                        total -= self._length_function(current_doc[0]) + (
                            separator_len if len(current_doc) > 1 else 0
                        )
                        current_doc = current_doc[1:]
            current_doc.append(d)
            total += _len + (separator_len if len(current_doc) > 1 else 0)
        doc = self._join_docs(current_doc, separator)
        if doc is not None:
            docs.append(doc)
        return docs


class CharacterTextSplitter(TextSplitter):
    """Splitting text that looks at characters."""

    def __init__(
        self, separator: str = "\n\n", is_separator_regex: bool = False, **kwargs: Any
    ) -> None:
        """Create a new TextSplitter."""
        super().__init__(**kwargs)
        self._separator = separator
        self._is_separator_regex = is_separator_regex

    def split_text(self, text: str) -> List[str]:
        """Split incoming text and return chunks."""
        # First we naively split the large input into a bunch of smaller ones.
        separator = (
            self._separator if self._is_separator_regex else re.escape(
                self._separator)
        )
        splits = split_text_with_regex(text, separator, self._keep_separator)
        _separator = "" if self._keep_separator else self._separator
        return self._merge_splits(splits, _separator)


def split_text_with_regex(
    text: str, separator: str, keep_separator: bool
) -> List[str]:
    # Now that we have the separator, split the text
    if separator:
        if keep_separator:
            # The parentheses in the pattern keep the delimiters in the result.
            _splits = re.split(f"({separator})", text)
            splits = [_splits[i] + _splits[i + 1]
                      for i in range(1, len(_splits), 2)]
            if len(_splits) % 2 == 0:
                splits += _splits[-1:]
            splits = [_splits[0]] + splits
        else:
            splits = re.split(separator, text)
    else:
        splits = list(text)
    return [s for s in splits if s != ""]



from openai import (
    APIConnectionError as OAIAPIConnectionError,
    AuthenticationError as OAIAuthenticationError,
    BadRequestError as OAIBadRequestError,
    RateLimitError as OAIRateLimitError,
    APIStatusError as OAIAPIStatusError,
    OpenAIError as OAIOpenAIError,
    APIError as OAIAPIError,
    APITimeoutError as OAIAPITimeoutError,
    APIResponseValidationError as OAIAPIResponseValidationError
)
import httpx


class AuthenticationError(OAIAuthenticationError):  # type: ignore
    def __init__(self, message, llm_provider, model, response: httpx.Response):
        self.status_code = 401
        self.message = message
        self.llm_provider = llm_provider
        self.model = model
        super().__init__(
            self.message,
            response=response,
            body=None
        )  # Call the base class constructor with the parameters it needs


class BadRequestError(OAIBadRequestError):  # type: ignore
    def __init__(self, message, model, llm_provider, response: httpx.Response):
        self.status_code = 400
        self.message = message
        self.model = model
        self.llm_provider = llm_provider
        super().__init__(
            self.message,
            response=response,
            body=None
        )  # Call the base class constructor with the parameters it needs


class Timeout(OAIAPITimeoutError):  # type: ignore
    def __init__(self, message, model, llm_provider):
        self.status_code = 408
        self.message = message
        self.model = model
        self.llm_provider = llm_provider
        request = httpx.Request(method="POST", url="https://api.openai.com/v1")
        super().__init__(
            request=request
        )  # Call the base class constructor with the parameters it needs


class RateLimitError(OAIRateLimitError):  # type: ignore
    def __init__(self, message, llm_provider, model, response: httpx.Response):
        self.status_code = 429
        self.message = message
        self.llm_provider = llm_provider
        self.modle = model
        super().__init__(
            self.message,
            response=response,
            body=None
        )  


class ContextWindowExceededError(OAIBadRequestError):
    def __init__(self, message, model, llm_provider, response: httpx.Response):
        self.status_code = 400
        self.message = message
        self.model = model
        self.llm_provider = llm_provider
        super().__init__(
            message=self.message,
            model=self.model,  # type: ignore
            llm_provider=self.llm_provider,  # type: ignore
            response=response
        )

class ServiceUnavailableError(OAIAPIStatusError):  # type: ignore
    def __init__(self, message, llm_provider, model, response: httpx.Response):
        self.status_code = 503
        self.message = message
        self.llm_provider = llm_provider
        self.model = model
        super().__init__(
            self.message,
            response=response,
            body=None
        )  # Call the base class constructor with the parameters it needs


class APIError(OAIAPIError):
    def __init__(self, status_code, message, llm_provider, model, request: httpx.Request):
        self.status_code = status_code
        self.message = message
        self.llm_provider = llm_provider
        self.model = model
        super().__init__(
            self.message,
            request=request,  # type: ignore
            body=None
        )

# raised if an invalid request (not get, delete, put, post) is made


class APIConnectionError(OAIAPIConnectionError): 
  def __init__(self, message, llm_provider, model, request: httpx.Request):
        self.message = message
        self.llm_provider = llm_provider
        self.model = model
        self.status_code = 500
        super().__init__(
            message=self.message,
            request=request
        )

# raised if an invalid request (not get, delete, put, post) is made


class APIResponseValidationError(OAIAPIResponseValidationError):  
    def __init__(self, message, llm_provider, model):
        self.message = message
        self.llm_provider = llm_provider
        self.model = model
        request = httpx.Request(method="POST", url="https://api.openai.com/v1")
        response = httpx.Response(status_code=500, request=request)
        super().__init__(
            response=response,
            body=None,
            message=message
        )


class OpenAIError(OAIOpenAIError):  
    def __init__(self, original_exception):
        self.status_code = original_exception.http_status
        super().__init__(
            http_body=original_exception.http_body, # type: ignore
            http_status=original_exception.http_status, # type: ignore
            json_body=original_exception.json_body, # type: ignore
            headers=original_exception.headers, # type: ignore
            code=original_exception.code, # type: ignore
        )
        self.llm_provider = "openai"


class BudgetExceededError(Exception):
    def __init__(self, current_cost, max_budget):
        self.current_cost = current_cost
        self.max_budget = max_budget
        message = f"Budget has been exceeded! Current cost: {current_cost}, Max budget: {max_budget}"
        super().__init__(message)


class UnsupportedParamsError(Exception):
    def __init__(self, status_code, message):
        self.status_code = status_code
        self.message = message
        super().__init__(
            self.message
        )  # Call the base class constructor with the parameters it needs



import re
from typing import Any, Callable
import json


def parse_partial_json(s: str, *, strict: bool = False) -> Any:
    try:
        return json.loads(s, strict=strict)
    except json.JSONDecodeError:
        pass

    new_s = ""
    stack = []
    is_inside_string = False
    escaped = False

    # Process each character in the string one at a time.
    for char in s:
        if is_inside_string:
            if char == '"' and not escaped:
                is_inside_string = False
            elif char == "\n" and not escaped:
                # Replace the newline character with the escape sequence.
                char = "\\n"
            elif char == "\\":
                escaped = not escaped
            else:
                escaped = False
        else:
            if char == '"':
                is_inside_string = True
                escaped = False
            elif char == "{":
                stack.append("}")
            elif char == "[":
                stack.append("]")
            elif char == "}" or char == "]":
                if stack and stack[-1] == char:
                    stack.pop()
                else:
                    # Mismatched closing character; the input is malformed.
                    return None

        new_s += char

    if is_inside_string:
        new_s += '"'

    for closing_char in reversed(stack):
        new_s += closing_char
    try:
        return json.loads(new_s, strict=strict)
    except json.JSONDecodeError:
        return None




def extract_json(text: str, parser: Callable[[str], Any] = json.loads):
    start_braces = text.find('{')
    end_braces = text.rfind('}')
    start_brackets = text.find('[')
    end_brackets = text.rfind(']')

    diff = text[start_brackets+1:start_braces].replace("\n", "").strip()
    if start_brackets != -1 and start_brackets < start_braces and diff == "":
      if end_brackets != -1:
        json_text = text[start_brackets:end_brackets+1]
        try:
            return parser(json_text)
        except json.JSONDecodeError:
            pass

    if start_braces != -1 and end_braces != -1:
        json_text = text[start_braces:end_braces+1]
        try:
            return parser(json_text)
        except json.JSONDecodeError:
            pass
    return None


def parse_json_markdown(json_string: str, *, parser: Callable[[str], Any] = json.loads) -> dict:
    match = re.search(r"```(json)?(.*)```", json_string, re.DOTALL)
    json_str = match.group(2) if match else json_string
    json_str = json_str.strip()
    # replace escaped newlines with actual newlines
    json_str = re.sub(r'\\n', r'\n', json_str)
    # replace escaped tabs with actual tabs
    json_str = re.sub(r'\\t', r'\t', json_str)
    # replace escaped quotes with actual quotes
    json_str = re.sub(r'\\"', r'"', json_str)
    parsed = parser(json_str)
    return parsed


def extract_json_from_string(text: str):
    json_str = parse_json_markdown(text, parser=parse_partial_json)
    if json_str is None:
      json_str = extract_json(text, parser=parse_partial_json)
    if  isinstance(json_str, list) and len(json_str) > 0:
        json_str = json_str[0]
    return json_str




def completion_price_calculator(cost_prompt: float, cost_completion: float, prompt_token: int, completion_token: int):
    cost_inp = cost_prompt / 1000000 * prompt_token
    cost_out = cost_completion / 1000000 * completion_token
    return round(cost_inp + cost_out, 7)


import base64
import logging
import time

import numpy as np

# Create a logger object
logger: logging.Logger = logging.getLogger(__name__)

# Set the logging level
logger.setLevel(logging.INFO)

# Create a file handler
file_handler = logging.FileHandler('logs.log')
console_handler = logging.StreamHandler()

# Set the logging format
formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
console_handler.setFormatter(formatter)
file_handler.setFormatter(formatter)

# Add the handler to the logger
logger.addHandler(console_handler)
logger.addHandler(file_handler)


def trim_string(string, count: int) -> str:
    words = string.split()
    trimmed_words = words[:count]
    trimmed_string = ' '.join(trimmed_words)
    return trimmed_string

def run_with_time(func):
    start_time = time.time()

    func()

    end_time = time.time()
    execution_time = end_time - start_time
    logger.info(execution_time)


def mp3_to_base64(file_path):
    with open(file_path, 'rb') as binary_file:
        binary_data = binary_file.read()
        base64_data = base64.b64encode(binary_data)
        base64_message = base64_data.decode('utf-8')
    return base64_message


def numpy_to_base64(array: np.ndarray) -> str:
    return base64.b64encode(array).decode('utf-8')


def base64_to_numpy(base64_str: str) -> np.ndarray:
    return np.frombuffer(base64.b64decode(base64_str), dtype=np.float32)



from typing import List, Optional, Union
from tenacity import RetryCallState

from .base import FeaturesState
from .response_types import (
    CompletionData, 
    ChatData,
    EmbeddingsData,
    ImageData,
    TTSData,
    TranscriptionData,
)

class CallbackHandler:

    def __init__(self, features_state: FeaturesState = FeaturesState(), **kwargs):
        self.features_state = features_state
        pass

    def on_llm_start(self, data: CompletionData, **kwargs):
        pass

    def on_llm_end(self, data: CompletionData, **kwargs):
        pass

    def on_chat_start(self, data: ChatData, **kwargs):
        pass

    def on_chat_end(self, data: ChatData, **kwargs):
        pass

    def on_llm_response(self, data: Union[CompletionData, ChatData], **kwargs):
        pass

    def on_llm_new_token(self, token: str, **kwargs):
        pass

    def on_embeddings_start(self,  data: EmbeddingsData, **kwargs):
        pass

    def on_embeddings_end(self,  data: EmbeddingsData, **kwargs):
        pass

    def on_image_start(self,  data: ImageData, **kwargs):
        pass

    def on_image_end(self,  data: ImageData, **kwargs):
        pass

    def on_tts_start(self,  data: TTSData, **kwargs):
        pass

    def on_tts_end(self,  data: TTSData, **kwargs):
        pass

    def on_transcription_start(self,  data: TranscriptionData, **kwargs):
        pass

    def on_transcription_end(self,  data: TranscriptionData, **kwargs):
        pass

    def on_retry(self, retry_state: RetryCallState, **kwargs):
        pass

Callbacks = Optional[List[CallbackHandler]]

def callback_from_feature(feature: str, callbacks: List[CallbackHandler]):
    res: List[CallbackHandler] = []
    for callback in callbacks:
        if callback.features_state.model_dump()[feature] == True:
            res.append(callback)
    return res

class CallbackManager:
    def __init__(self, handlers: List[CallbackHandler], features_state: FeaturesState = FeaturesState(), **kwargs):
        self.features_state = features_state
        self.handlers = handlers

    def on_llm_start(self, data: CompletionData, callbacks: Callbacks, **kwargs):
        if self.features_state.on_llm_start == False:
            return
        handlers = callback_from_feature('on_llm_start', self.handlers + (callbacks or []))
        for handler in handlers:
            handler.on_llm_start(data, kwargs=kwargs)

    def on_llm_end(self, data: CompletionData, callbacks: Callbacks, **kwargs):
        if self.features_state.on_llm_end == False:
            return
        handlers = callback_from_feature(
            'on_llm_end', self.handlers + (callbacks or []))
        for handler in handlers:
            handler.on_llm_end(data, kwargs=kwargs)

    def on_chat_start(self, data: ChatData, callbacks: Callbacks, **kwargs):
        if self.features_state.on_chat_start == False:
            return
        handlers = callback_from_feature(
            'on_chat_start', self.handlers + (callbacks or []))
        for handler in handlers:
            handler.on_chat_start(data, kwargs=kwargs)

    def on_chat_end(self, data: ChatData, callbacks: Callbacks, **kwargs):
        if self.features_state.on_chat_end == False:
            return
        handlers = callback_from_feature(
            'on_chat_end', self.handlers + (callbacks or []))
        for handler in handlers:
            handler.on_chat_end(data, kwargs=kwargs)

    def on_llm_response(self, data: Union[CompletionData, ChatData], callbacks: Callbacks, **kwargs):
        if self.features_state.on_llm_response == False:
            return
        handlers = callback_from_feature(
            'on_llm_response', self.handlers + (callbacks or []))
        for handler in handlers:
            handler.on_llm_response(data, kwargs=kwargs)

    def on_llm_new_token(self, token: str, callbacks: Callbacks, **kwargs):
        if self.features_state.on_llm_new_token == False:
            return
        handlers = callback_from_feature(
            'on_llm_new_token', self.handlers + (callbacks or []))
        for handler in handlers:
            handler.on_llm_new_token(token, kwargs=kwargs)

    def on_embeddings_start(self, data: EmbeddingsData, callbacks: Callbacks, **kwargs):
        if self.features_state.on_embeddings_start == False:
            return
        handlers = callback_from_feature(
            'on_embeddings_start', self.handlers + (callbacks or []))
        for handler in handlers:
            handler.on_embeddings_start(data, kwargs=kwargs)

    def on_embeddings_end(self, data: EmbeddingsData, callbacks: Callbacks, **kwargs):
        if self.features_state.on_embeddings_end == False:
            return
        handlers = callback_from_feature(
            'on_embeddings_end', self.handlers + (callbacks or []))
        for handler in handlers:
            handler.on_embeddings_end(data, kwargs=kwargs)

    def on_image_start(self, data: ImageData, callbacks: Callbacks, **kwargs):
        if self.features_state.on_image_start == False:
            return
        handlers = callback_from_feature(
            'on_image_start', self.handlers + (callbacks or []))
        for handler in handlers:
            handler.on_image_start(data, kwargs=kwargs)

    def on_image_end(self, data: ImageData, callbacks: Callbacks, **kwargs):
        if self.features_state.on_image_end == False:
            return
        handlers = callback_from_feature(
            'on_image_end', self.handlers + (callbacks or []))
        for handler in handlers:
            handler.on_image_end(data, kwargs=kwargs)

    def on_tts_start(self, data: TTSData, callbacks: Callbacks, **kwargs):
        if self.features_state.on_tts_start == False:
            return
        handlers = callback_from_feature(
            'on_tts_start', self.handlers + (callbacks or []))
        for handler in handlers:
            handler.on_tts_start(data, kwargs=kwargs)

    def on_tts_end(self, data: TTSData, callbacks: Callbacks, **kwargs):
        if self.features_state.on_tts_end == False:
            return
        handlers = callback_from_feature(
            'on_tts_end', self.handlers + (callbacks or []))
        for handler in handlers:
            handler.on_tts_end(data, kwargs=kwargs)

    def on_transcription_start(self, data: TranscriptionData, callbacks: Callbacks, **kwargs):
        if self.features_state.on_transcription_start == False:
            return
        handlers = callback_from_feature(
            'on_transcription_start', self.handlers + (callbacks or []))
        for handler in handlers:
            handler.on_transcription_start(data, kwargs=kwargs)

    def on_transcription_end(self, data: TranscriptionData, callbacks: Callbacks, **kwargs):
        if self.features_state.on_transcription_end == False:
            return
        handlers = callback_from_feature(
            'on_transcription_end', self.handlers + (callbacks or []))
        for handler in handlers:
            handler.on_transcription_end(data, kwargs=kwargs)

    def on_retry(self, retry_state: RetryCallState, callbacks: Callbacks, **kwargs):
        if self.features_state.on_retry == False:
            return
        handlers = callback_from_feature('on_retry', self.handlers + (callbacks or []))
        for handler in handlers:
            handler.on_retry(retry_state, kwargs=kwargs)



import functools
import random
import string
import time
from typing import Any, Callable, Dict, List, Literal, Optional, Type, Union

import numpy as np
import httpx

from .utils import num_tokens_from_string
from .provider_inputs import BaseProviderInput
from .manipulations import ManipulationManager, Manipulations
from .callbacks import CallbackManager, Callbacks
from .response_types import (
    BaseMessage,
    ChatChoices,
    ChatContent,
    ChatResponse,
    CompletionChoices,
    EmbeddingsData,
    FunctionCallContent,
    ImageData,
    Message,
    ChatMessages,
    CompletionData,
    ChatData,
    TTSData,
    TranscriptionData,
    Usage,
)


class ProviderBase:
    _client_session: Optional[httpx.Client] = None
    _aclient_session: Optional[httpx.AsyncClient] = None

    class Config:
        arbitrary_types_allowed = True

    configs: BaseProviderInput
    callback_manager: CallbackManager
    manipulation_manager: Optional[ManipulationManager] = None

    def __init__(self, configs: BaseProviderInput, callback_manager: CallbackManager, manipulation_manager: Optional[ManipulationManager] = None):
        self.manipulation_manager = manipulation_manager
        self.callback_manager = callback_manager
        self.configs = configs

    def create_client_session(self):
        if samosila.client_session:
            _client_session = samosila.client_session
        else:
            _client_session = httpx.Client()

        return _client_session

    def create_aclient_session(self):
        if samosila.aclient_session:
            _aclient_session = samosila.aclient_session
        else:
            _aclient_session = httpx.AsyncClient()

        return _aclient_session

    def __exit__(self):
        if self._client_session is not None:
            self._client_session.close()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._aclient_session is not None:
            await self._aclient_session.aclose()

    def completion(
        self,
        prompt: Message,
        data: Optional[CompletionData] = None,
        manipulations: Manipulations = None,
        callbacks: Callbacks = None,
        metadata: Dict[str, Any] = {},
    ) -> CompletionData:
        raise NotImplementedError()

    def chat(
        self,
        messages: ChatMessages,
        data: Optional[ChatData] = None,
        manipulations: Manipulations = None,
        callbacks: Callbacks = None,
        metadata: Dict[str, Any] = {},
    ) -> ChatData:
        raise NotImplementedError()

    def embedding(
        self,
        texts: List[str],
        data: Optional[EmbeddingsData] = None,
        manipulations: Manipulations = None,
        callbacks: Callbacks = None,
        metadata: Dict[str, Any] = {},
    ) -> EmbeddingsData:
        raise NotImplementedError()

    def image_generation(
        self,
        prompt: str,
        size: str,
        data: Optional[ImageData] = None,
        negative_prompt: Optional[str] = None,
        steps: Optional[int] = None,
        manipulations: Manipulations = None,
        callbacks: Callbacks = None,
        metadata: Dict[str, Any] = {},
    ) -> ImageData:
        raise NotImplementedError()

    def transcription(
        self,
        audio: str,
        data: Optional[TranscriptionData] = None,
        input_type: Literal["b64_json", "url"] = "b64_json",
        manipulations: Manipulations = None,
        callbacks: Callbacks = None,
        metadata: Dict[str, Any] = {},
    ) -> TranscriptionData:
        raise NotImplementedError()

    def data_to_response(
        self,
        data: Union[CompletionData, ChatData],
        text: str,
        model: Optional[str] = None,
        uuid: Optional[str] = None,
        created: Optional[int] = None,
        usage: Optional[Usage] = None,
        function: Optional[FunctionCallContent] = None,
    ):

        data.provider_model = model or self.configs.provider_model
        data.uuid = uuid or data.metadata.get("run_id")
        data.created = created or int(time.time())
        data.usage = usage
        if data.usage is None:
            prompt_tokens = num_tokens_from_string(data.prompt.content if isinstance(
                data, CompletionData) else "".join([x.content for x in data.messages]))
            completion_tokens = num_tokens_from_string(text)
            data.usage = Usage(
                completion_tokens=completion_tokens,
                prompt_tokens=prompt_tokens,
                total_tokens=completion_tokens + prompt_tokens,
            )
        if isinstance(data, ChatResponse):
            data.response = [
                ChatChoices(
                    index=0,
                    finish_reason='stop',
                    message=ChatContent(
                        content=text,
                        role='assistant',
                        function_call=function
                    )
                )
            ]
        else:
            data.response = [
                CompletionChoices(
                    text=text,
                    finish_reason='stop',
                    index=0,
                )
            ]
        return data


completion_type = Callable[[Any, Message, Optional[CompletionData],
                            Manipulations, Callbacks, Dict[str, Any]], CompletionData | None]

chat_type = Callable[[Any, ChatMessages, Optional[ChatData],
                      Manipulations, Callbacks, Dict[str, Any]], ChatData | None]

embeddings_type = Callable[[Any, List[str], Optional[EmbeddingsData],
                            Manipulations, Callbacks, Dict[str, Any]], EmbeddingsData | None]

image_generation_type = Callable[[Any, str, Literal['256x256', '512x512', '1024x1024', '1792x1024', '1024x1792'],
                                  Optional[ImageData], Optional[str], Optional[int],
                                  Manipulations, Callbacks, Dict[str, Any]], ImageData | None]


def completion_decorator(func):
    @functools.wraps(func)
    def wrapper(
        self: Type[ProviderBase],
        prompt: Message,
        data: Optional[CompletionData] = None,
        manipulations: Manipulations = None,
        callbacks: Callbacks = None,
        metadata: Dict[str, Any] = {},
        **kwargs
    ):
        if not issubclass(type(self), ProviderBase):
            raise TypeError("Class must extend ProviderBase")

        metadata["run_id"] = "opfr-" + "".join(random.choices(
            string.ascii_letters + string.digits, k=28))

        if isinstance(prompt, str):
            prompt = BaseMessage(content=prompt, role="user")
        if data is None:
            data = CompletionData(
                prompt=prompt,
                provider_model=self.configs.provider_model,
                metadata=metadata,
            )

        if self.manipulation_manager:
            data = self.manipulation_manager.on_llm_start(
                data, manipulations)
        self.callback_manager.on_llm_start(data, callbacks)

        if self.configs.cache == False or data.response is None:
            data = func(self, prompt, data, manipulations,
                        callbacks, metadata)
        else:
            data.metadata["caching"] = True

        if self.manipulation_manager:
            data = self.manipulation_manager.on_llm_end(
                data, manipulations)
        self.callback_manager.on_llm_end(data, callbacks)
        return data
    return wrapper


def chat_decorator(func):
    @functools.wraps(func)
    def wrapper(
        self: Type[ProviderBase],
        messages: ChatMessages,
        data: Optional[ChatData] = None,
        manipulations: Manipulations = None,
        callbacks: Callbacks = None,
        metadata: Dict[str, Any] = {},
    ):
        metadata["run_id"] = "opfr-" + "".join(random.choices(
            string.ascii_letters + string.digits, k=28))

        _messages = []
        for message in messages:
            if isinstance(message, Dict):
                message = BaseMessage(
                    content=message["content"],
                    role=message["type"] or message["role"]
                )
            _messages.append(message)
        if data is None:
            data = ChatData(
                messages=_messages,
                provider_model=self.configs.provider_model,
                metadata=metadata,
            )
        if self.manipulation_manager:
            data = self.manipulation_manager.on_chat_start(
                data, manipulations)
        self.callback_manager.on_chat_start(data, callbacks)

        if self.configs.cache == False or data.response is None:
            data = func(self, messages, data,
                        manipulations, callbacks, metadata)
        else:
            data.metadata["caching"] = True

        if self.manipulation_manager:
            data = self.manipulation_manager.on_chat_end(
                data, manipulations)
        self.callback_manager.on_chat_end(data, callbacks)
        return data
    return wrapper


def embeddings_decorator(func):
    @functools.wraps(func)
    def wrapper(
        self: Type[ProviderBase],
        texts: List[str],
        data: Optional[EmbeddingsData] = None,
        manipulations: Manipulations = None,
        callbacks: Callbacks = None,
        metadata: Dict[str, Any] = {},
    ):
        metadata["run_id"] = "opfr-" + "".join(random.choices(
            string.ascii_letters + string.digits, k=28))

        if data is None:
            data = EmbeddingsData(
                embeddings_inputs=texts,
                provider_model=self.configs.provider_model,
                metadata=metadata,
            )

        if self.manipulation_manager:
            data = self.manipulation_manager.on_embeddings_start(
                data, manipulations)
        self.callback_manager.on_embeddings_start(data, callbacks)

        if self.configs.cache == False or data.response is None:
            data = func(self, texts, data, manipulations, callbacks, metadata)
        else:
            data.metadata["caching"] = True

        if self.manipulation_manager:
            data = self.manipulation_manager.on_embeddings_end(
                data, manipulations)
        self.callback_manager.on_embeddings_end(data, callbacks)
        return data

    return wrapper


def image_generation_decorator(func):
    @functools.wraps(func)
    def wrapper(
        self: Type[ProviderBase],
        prompt: str,
        size: Literal['256x256', '512x512', '1024x1024', '1792x1024', '1024x1792'],
        data: Optional[ImageData] = None,
        negative_prompt: Optional[str] = None,
        steps: Optional[int] = None,
        manipulations: Manipulations = None,
        callbacks: Callbacks = None,
        metadata: Dict[str, Any] = {},
    ):
        metadata["run_id"] = "opfr-" + "".join(random.choices(
            string.ascii_letters + string.digits, k=28))

        if data is None:
            data = ImageData(
                prompt=prompt,
                size=size,
                negative_prompt=negative_prompt,
                steps=steps,
                provider_model=self.configs.provider_model,
                response_format=metadata.get(
                    "response_format") or "b64_json",
                metadata=metadata,
            )

        if self.manipulation_manager:
            data = self.manipulation_manager.on_image_start(
                data, manipulations)
        self.callback_manager.on_image_start(data, callbacks)

        if self.configs.cache == False or data.response is None:
            data = func(self, prompt, size, data, negative_prompt,
                        steps, manipulations, callbacks, metadata)
        else:
            data.metadata["caching"] = True

        if self.manipulation_manager:
            data = self.manipulation_manager.on_image_end(
                data, manipulations)
        self.callback_manager.on_image_end(data, callbacks)
        return data

    return wrapper


def tts_decorator(func):
    @functools.wraps(func)
    def wrapper(
        self: Type[ProviderBase],
        text: str,
        data: Optional[TTSData] = None,
        gender: Literal["Male", "Female"] = "Male",
        file_name: Optional[str] = None,
        manipulations: Manipulations = None,
        callbacks: Callbacks = None,
        metadata: Dict[str, Any] = {},
    ):
        metadata["run_id"] = "opfr-" + "".join(random.choices(
            string.ascii_letters + string.digits, k=28))

        if data is None:
            data = TTSData(
                text=text,
                download_path=file_name,
                gender=gender,
                provider_model=self.configs.provider_model,
                response_format=metadata.get(
                    "response_format") or "b64_json",
                metadata=metadata,
            )

        if self.manipulation_manager:
            data = self.manipulation_manager.on_tts_start(
                data, manipulations)
        self.callback_manager.on_tts_start(data, callbacks)

        if self.configs.cache == False or data.response is None:
            data = func(self, text, data, gender, file_name,
                        manipulations, callbacks, metadata)
        else:
            data.metadata["caching"] = True

        if self.manipulation_manager:
            data = self.manipulation_manager.on_tts_end(
                data, manipulations)
        self.callback_manager.on_tts_end(data, callbacks)
        return data

    return wrapper


def transcription_decorator(func):
    @functools.wraps(func)
    def wrapper(
        self: Type[ProviderBase],
        audio: str | np.ndarray,
        data: Optional[TranscriptionData] = None,
        input_type: Literal["b64_json", "url", "numpy"] = "b64_json",
        manipulations: Manipulations = None,
        callbacks: Callbacks = None,
        metadata: Dict[str, Any] = {},
    ):
        metadata["run_id"] = "opfr-" + "".join(random.choices(
            string.ascii_letters + string.digits, k=28))

        if data is None:
            data = TranscriptionData(
                audio=audio,
                input_type=input_type,
                provider_model=self.configs.provider_model,
                metadata=metadata,
            )

        if self.manipulation_manager:
            data = self.manipulation_manager.on_transcription_start(
                data, manipulations)
        self.callback_manager.on_transcription_start(data, callbacks)

        if self.configs.cache == False or data.response is None:
            data = func(self, audio, data, input_type,
                        manipulations, callbacks, metadata)
        else:
            data.metadata["caching"] = True

        if self.manipulation_manager:
            data = self.manipulation_manager.on_transcription_end(
                data, manipulations)
        self.callback_manager.on_transcription_end(data, callbacks)
        return data

    return wrapper



from enum import Enum
from typing import Any, ClassVar, Dict, List, Optional
from pydantic import BaseModel, model_validator


class BaseProviderInput(BaseModel):
    provider_model: str
    cache: bool = False
    verbose: bool = True
    max_retries: int = 4
    streaming: bool = False
    provider_kwargs: Dict[str, Any] = {}

    @model_validator(mode="before")
    @classmethod
    def validate_environment(cls, values: Dict) -> Dict:
        if values.get("max_tokens") is not None and values["max_tokens"] <= 0:
            raise ValueError("max_tokens must be positive")

        if values.get("temperature") is not None and not 0 <= values["temperature"] <= 1:
            raise ValueError("temperature must be in the range [0.0, 1.0]")

        if values.get("top_p") is not None and not 0 <= values["top_p"] <= 1:
            raise ValueError("top_p must be in the range [0.0, 1.0]")

        if values.get("top_k") is not None and values["top_k"] <= 0:
            raise ValueError("top_k must be positive")

        return values


class FireworksInput(BaseProviderInput):
    provider_name: ClassVar[str] = 'fireworks'
    api_key: str | None = None


class HuggingFaceInput(BaseProviderInput):
    provider_name: ClassVar[str] = 'huggingface'
    class_type: Optional[str] = None
    multi_process: bool = False
    cache_folder: Optional[str] = None
    device: str = "cpu"


class CTranslateInput(BaseProviderInput):
    provider_name: ClassVar[str] = 'ctranslate2'
    device: str = "cpu"


class VoyageInput(BaseProviderInput):
    provider_name: ClassVar[str] = 'voyage'
    api_key: str


class GradientInput(BaseProviderInput):
    provider_name: ClassVar[str] = 'gradient'
    api_key: str
    workspace_id: Optional[str] = None


class NovitaInput(BaseProviderInput):
    provider_name: ClassVar[str] = 'novita'
    api_key: str
    sampler_name: str = "DPM++ 2M Karras"
    n: int = 1
    cfg: float = 6

class SegmindInput(BaseProviderInput):
    provider_name: ClassVar[str] = 'segmind'
    api_key: str
    sampler_name: str = "DPM++ 2M Karras"
    n: int = 1
    cfg: float = 6


class GoogleGenAIInput(BaseProviderInput):
    provider_name: ClassVar[str] = 'generativeai'
    api_key: str
    top_k: Optional[int] = None
    top_p: Optional[float] = None
    temperature: Optional[float] = None


class CohereInput(BaseProviderInput):
    provider_name: ClassVar[str] = 'cohere'
    api_key: str
    top_k: Optional[int] = None
    top_p: Optional[float] = None
    temperature: Optional[float] = None
    max_tokens: Optional[int] = None


class OpenAIInput(BaseProviderInput):
    provider_name: ClassVar[str] = 'openai'
    api_key: str
    top_k: Optional[int] = None
    top_p: Optional[float] = None
    temperature: Optional[float] = None
    frequency_penalty: Optional[float] = None
    presence_penalty: Optional[float] = None
    max_tokens: Optional[int] = None
    base_url: Optional[str] = None
    stop: Optional[List[str]] = None
    base_url: Optional[str] = None
    n: int = 1


class TogetherInput(BaseProviderInput):
    provider_name: ClassVar[str] = 'together'
    api_key: str
    top_k: Optional[int] = None
    top_p: Optional[float] = None
    temperature: Optional[float] = None
    repeat_penalty: Optional[float] = None
    max_tokens: Optional[int] = None
    n: int = 1


class TTSInput(BaseProviderInput):
    provider_name: ClassVar[str] = 'tts'
    emotion: Optional[str] = "Neutral"
    api_key: Optional[str] = None


class FasterWhisperInput(BaseProviderInput):
    provider_name: ClassVar[str] = 'faster-whisper'
    emotion: Optional[str] = "Neutral"
    api_key: Optional[str] = None


class LlamacppInput(BaseProviderInput):
    provider_name: ClassVar[str] = 'llamacpp'
    top_k: Optional[int] = None
    chat_format: str = 'llama-2'
    top_p: Optional[float] = None
    temperature: Optional[float] = None
    n_ctx: Optional[int] = None
    stop: Optional[List[str]] = None
    max_tokens: Optional[int] = None
    repeat_penalty: Optional[float] = None
    grammer: str | None = None
    f16_kv: bool = True
    n_gpu_layers: int = 0
    n_threads: int = 4


class CompleteInput(LlamacppInput, TogetherInput, OpenAIInput, CohereInput, CTranslateInput, SegmindInput,
        NovitaInput, FireworksInput, GoogleGenAIInput, VoyageInput, HuggingFaceInput, TTSInput, FasterWhisperInput):
    provider_name: ClassVar[str] = 'random'


class ProviderType(Enum):
    AI21 = "ai21"
    COHERE = "cohere"
    FAKE = "fake"
    FIREWORKS = "fireworks"
    FREE = "free"
    LLAMACPP = "llamacpp"
    OPENAI = "openai"
    PALM = "palm"
    TOGETHER = "together"

    @classmethod
    def get_type(cls, type: str):
        type_enum_value = None
        for enum_value in ProviderType:
            if type == enum_value.value:
                type_enum_value = enum_value
                break
        return type_enum_value or cls.FREE




import numpy as np
from abc import ABC
from typing import Any, Dict, List, Literal, Optional, Union
from pydantic import BaseModel


class BaseMessage(BaseModel, ABC):
    content: str
    role: str


class SystemMessage(BaseMessage):
    role: str = "system"


class HumanMessage(BaseMessage):
    role: str = "human"


class UserMessage(BaseMessage):
    role: str = "user"


class AIMessage(BaseMessage):
    role: str = "ai"


Message = Union[BaseMessage, str]
ChatMessages = Union[List[BaseMessage], List[Dict[str, Any]]]


class Function(BaseModel):
    name: str 
    description: str 
    parameters: Dict[str, Any]


class Usage(BaseModel):
    completion_tokens: Optional[int] = None
    prompt_tokens: Optional[int] = None
    total_tokens: Optional[int] = None
    cost: Optional[int] = None


class CompletionChoices(BaseModel):
    finish_reason: str
    index: int
    text: str


class FunctionCallContent(BaseModel):
    name: str
    arguments: str


class ChatContent(BaseModel):
    content: Optional[str] = None
    function_call: Optional[FunctionCallContent] = None
    role: Optional[str] = None

class ImageContent(BaseModel):
    url: Optional[str] = None
    b64_json: Optional[str] = None

class ChatChoices(BaseModel):
    finish_reason: str
    index: int
    message: ChatContent

class VoiceSegments(BaseModel):
    start: float
    end: float
    content: str


class ChatCompletionRequest(BaseModel):
    provider_model: Optional[str] = None
    messages: List[BaseMessage]
    functions: Optional[List[Function]] = None


class CompletionRequest(BaseModel):
    provider_model: Optional[str] = None
    prompt: BaseMessage


class EmbeddingsRequest(BaseModel):
    provider_model: str
    embeddings_inputs: List[str]


class ImageRequest(BaseModel):
    provider_model: str
    prompt: str
    size: Literal['256x256', '512x512', '1024x1024', '1792x1024', '1024x1792']
    negative_prompt: Optional[str] = None
    steps: Optional[int] = None
    response_format: Literal['url', 'b64_json'] = "b64_json"
    download_path: Optional[str] = None

class TTSRequest(BaseModel):
    provider_model: str
    text: str
    voice: Optional[str] = None
    target_lang: Optional[str] = 'en'
    response_format: Literal['url', 'b64_json'] = "b64_json"
    download_path: Optional[str] = None

class TranscriptionRequest(BaseModel):
    class Config:
        arbitrary_types_allowed = True
    
    provider_model: str
    audio: str | np.ndarray
    input_type: Literal["b64_json", "url", "numpy"]
    lang: Optional[str]

class CompletionResponse(BaseModel):
    response_llm: Optional[str] = None
    provider_model: Optional[str] = None
    uuid: Optional[str] = None
    object_type: Optional[str] = None
    created: Optional[int] = None
    usage: Optional[Usage] = None
    response: Optional[List[CompletionChoices]] = None


class EmbeddingsResponse(BaseModel):
    embeddings_data: Optional[List[List[float]]] = None
    provider_model: Optional[str] = None
    uuid: Optional[str] = None
    object_type: Optional[str] = None
    created: Optional[int] = None
    response: Optional[List[List[float]]] = None
   

class ChatResponse(BaseModel):
    messages_response: Optional[List[BaseMessage]] = None
    provider_model: Optional[str] = None
    uuid: Optional[str] = None
    object_type: Optional[str] = None
    created: Optional[int] = None
    usage: Optional[Usage] = None
    response: Optional[List[ChatChoices]] = None


class ImageResponse(BaseModel):
    provider_model: Optional[str] = None
    uuid: Optional[str] = None
    created: Optional[int] = None
    response: Optional[List[ImageContent]] = None


class TTSResponse(BaseModel):
    provider_model: Optional[str] = None
    uuid: Optional[str] = None
    gender: Literal["Male", "Female"] = "Male"
    created: Optional[int] = None
    response: Optional[str] = None


class TranscriptionResponse(BaseModel):
    provider_model: str
    uuid: Optional[str] = None
    created: Optional[int] = None
    response: Optional[str] = None
    segments: Optional[List[VoiceSegments]] = None
    lang: Optional[str] = None



class ChatData(ChatResponse, ChatCompletionRequest):
    provider: Optional[str] = None
    metadata: Dict[str, Any]


class CompletionData(CompletionResponse, CompletionRequest):
    provider: Optional[str] = None
    metadata: Dict[str, Any]


class EmbeddingsData(EmbeddingsResponse, EmbeddingsRequest):
    provider: Optional[str] = None
    metadata: Dict[str, Any]

class ImageData(ImageResponse, ImageRequest):
    provider: Optional[str] = None
    metadata: Dict[str, Any]

class TTSData(TTSResponse, TTSRequest):
    provider: Optional[str] = None
    metadata: Dict[str, Any]

class TranscriptionData(TranscriptionResponse, TranscriptionRequest):
    provider: Optional[str] = None
    metadata: Dict[str, Any]



import pathlib
from setuptools import setup, find_packages

setup(
    name="samosila-core",
    version="0.0.3",
    author="Mohammad Javad Hosseini",
    author_email="johndoe2561357@protonmail.com",
    description="",
    long_description=pathlib.Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    license="MIT",
    python_requires=">=3.10,<3.12",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "appdirs",
        "certifi",
        "cryptography",
        "regex",
        "requests",
        "requests-file",
        "requests-html",
        "tokenizers>=0.15",
        "Pillow>=10",
        "python-dotenv>=1",
        "pydantic_settings>=2",
        "six",
        "tiktoken>=0.5",
        "unstructured>=0.11",
        "validators",
        "httpx>=0.25",
        "pydantic>=2",
        "tenacity>=8.2.3",
        "openai>=1",
        "aiohttp"
    ],
    extras_require={
        "full": [
            "cohere",
            "context-python",
            "ctranslate2",
            "datasets",
            "deepeval",
            "edge-tts",
            "elevenlabs",
            "faster-whisper>=0.10",
            "fireworks-ai>=0.9",
            "faiss-cpu>=1.7",
            "g4f",
            "google-generativeai>=0.3",
            "gptcache>=0.1.43",
            "gradientai",
            "huggingface-hub>=0.19",
            "jsonschema",
            "langchain>=0.0.350",
            "librosa>=0.10",
            "llama-index",
            "llama_cpp_python",
            "llmlingua",
            "lunary",
            "nltk>=3.8",
            "novita_client>=0.4",
            "omegaconf>=2.3",
            "prompt-toolkit>=3",
            "promptlayer",
            "PyAudio",
            "pymilvus>=2",
            "sentence-transformers>=2.2",
            "sentencepiece",
            "sentry-sdk>=1",
            "soundfile",
            "timm>=0.4.12",
            "together",
            "torch>=2",
            "torchaudio>=2",
            "torchvision>=0.15",
            "transformers>=4.36",
            "TTS>=0.22",
            "voyageai",
            "chromadb>=0.4",
            "deeplake",
            "pinecone-client>=2.2",
            "qdrant-client",
            "redis>=4.3",
            "replicate>=0.21",
            "weaviate-client>=3.25",
            "ai21",
            "lancedb"
        ]
    }
)



from .utils import logger, mp3_to_base64, run_with_time, trim_string, numpy_to_base64, base64_to_numpy
from .image import base64_to_image, image_to_base64, read_image
from .json import extract_json, extract_json_from_string, parse_json_markdown
from .text_splitter import TextSplitter, split_text_with_regex,CharacterTextSplitter

__all__: list[str] = [
    'logger',
    'base64_to_image',
    'image_to_base64',
    'read_image',
    'extract_json_from_string',
    'extract_json',
    'parse_json_markdown',
    "TextSplitter",
    "split_text_with_regex",
    "CharacterTextSplitter",
    "mp3_to_base64",
    "trim_string",
    "run_with_time",
    "numpy_to_base64",
    "base64_to_numpy",
]
